@@ -12,6 +12,8 @@ import string
 import subprocess
 import sys
 import tempfile
+import inspect
+import types
 
 from pwnlib import atexit
 from pwnlib.context import context
@@ -28,6 +30,7 @@ def align(alignment, x):
     Rounds `x` up to nearest multiple of the `alignment`.
 
     Example:
+
       >>> [align(5, n) for n in range(15)]
       [0, 5, 5, 5, 5, 5, 10, 10, 10, 10, 10, 15, 15, 15, 15]
     """
@@ -40,6 +43,7 @@ def align_down(alignment, x):
     Rounds `x` down to nearest multiple of the `alignment`.
 
     Example:
+
         >>> [align_down(5, n) for n in range(15)]
         [0, 0, 0, 0, 0, 5, 5, 5, 5, 5, 10, 10, 10, 10, 10]
     """
@@ -52,6 +56,7 @@ def binary_ip(host):
     Resolve host and return IP as four byte string.
 
     Example:
+
         >>> binary_ip("127.0.0.1")
         b'\\x7f\\x00\\x00\\x01'
     """
@@ -69,6 +74,7 @@ def size(n, abbrev = 'B', si = False):
       abbrev(str): String appended to the size, defaults to ``'B'``.
 
     Example:
+
         >>> size(451)
         '451B'
         >>> size(1000)
@@ -114,6 +120,7 @@ def read(path, count=-1, skip=0):
     Open file, return content.
 
     Examples:
+
         >>> read('/proc/self/exe')[:4]
         b'\x7fELF'
     """
@@ -139,6 +146,7 @@ def which(name, all = False, path=None):
 
     Works as the system command ``which``; searches $PATH for ``name`` and
     returns a full path if found.
+    Tries all of the file extensions in $PATHEXT on Windows too.
 
     If `all` is :const:`True` the set of all found locations is returned, else
     the first occurrence or :const:`None` is returned.
@@ -160,26 +168,35 @@ def which(name, all = False, path=None):
     if os.path.sep in name:
         return name
 
-    isroot = os.getuid() == 0
+    if sys.platform == 'win32':
+        pathexts = os.environ.get('PATHEXT', '').split(os.pathsep)
+        isroot = False
+    else:
+        pathexts = []
+        isroot = os.getuid() == 0
+    pathexts = [''] + pathexts
     out = set()
     try:
         path = path or os.environ['PATH']
     except KeyError:
         log.exception('Environment variable $PATH is not set')
-    for p in path.split(os.pathsep):
-        p = os.path.join(p, name)
-        if os.access(p, os.X_OK):
-            st = os.stat(p)
-            if not stat.S_ISREG(st.st_mode):
-                continue
-            # work around this issue: https://bugs.python.org/issue9311
-            if isroot and not \
-              st.st_mode & (stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH):
-                continue
-            if all:
-                out.add(p)
-            else:
-                return p
+    for path_part in path.split(os.pathsep):
+        for ext in pathexts:
+            nameext = name + ext
+            p = os.path.join(path_part, nameext)
+            if os.access(p, os.X_OK):
+                st = os.stat(p)
+                if not stat.S_ISREG(st.st_mode):
+                    continue
+                # work around this issue: https://bugs.python.org/issue9311
+                if isroot and not \
+                st.st_mode & (stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH):
+                    continue
+                if all:
+                    out.add(p)
+                    break
+                else:
+                    return p
     if all:
         return out
     else:
@@ -229,6 +246,10 @@ def normalize_argv_env(argv, env, log, level=2):
         for k,v in env_items:
             if not isinstance(k, (bytes, six.text_type)):
                 log.error('Environment keys must be strings: %r' % k)
+            # Check if = is in the key, Required check since we sometimes call ctypes.execve directly
+            # https://github.com/python/cpython/blob/025995feadaeebeef5d808f2564f0fd65b704ea5/Modules/posixmodule.c#L6476
+            if b'=' in packing._encode(k):
+                log.error('Environment keys may not contain "=": %r' % (k))
             if not isinstance(v, (bytes, six.text_type)):
                 log.error('Environment values must be strings: %r=%r' % (k,v))
             k = packing._need_bytes(k, level, 0x80)  # ASCII text is okay
@@ -293,7 +314,11 @@ def run_in_new_terminal(command, terminal=None, args=None, kill_at_exit=True, pr
         elif 'STY' in os.environ and which('screen'):
             terminal = 'screen'
             args     = ['-t','pwntools-gdb','bash','-c']
-        elif 'TERM_PROGRAM' in os.environ:
+        elif 'TERM_PROGRAM' in os.environ and os.environ['TERM_PROGRAM'] == "iTerm.app" and which('osascript'):
+            # if we're on a mac, and using iTerm
+            terminal = "osascript"
+            args     = []
+        elif 'TERM_PROGRAM' in os.environ and which(os.environ['TERM_PROGRAM']):
             terminal = os.environ['TERM_PROGRAM']
             args     = []
         elif 'DISPLAY' in os.environ and which('x-terminal-emulator'):
@@ -352,7 +377,6 @@ def run_in_new_terminal(command, terminal=None, args=None, kill_at_exit=True, pr
                     args.extend(['wsl.exe', '-d', distro_name, 'bash', '-c'])
                 else:
                     args.extend(['bash.exe', '-c'])
-                
 
     if not terminal:
         log.error('Could not find a terminal binary to use. Set context.terminal to your terminal.')
@@ -382,7 +406,7 @@ def run_in_new_terminal(command, terminal=None, args=None, kill_at_exit=True, pr
 import os
 os.execve({argv0!r}, {argv!r}, os.environ)
 '''
-        script = script.format(executable=sys.executable,
+        script = script.format(executable='/bin/env ' * (' ' in sys.executable) + sys.executable,
                                argv=command,
                                argv0=which(command[0]))
         script = script.lstrip()
@@ -396,6 +420,26 @@ os.execve({argv0!r}, {argv!r}, os.environ)
           argv += [tmp.name]
 
 
+    # if we're on a Mac and use iTerm, we use `osascript` to split the current window
+    # `command` was sanitized on the previous step. It is now either a string, or was written to a tmp file
+    # we run the command, which is now `argv[-1]`
+    if terminal == 'osascript':
+        osa_script = """
+tell application "iTerm"
+    tell current session of current window
+        set newSession to (split horizontally with default profile)
+    end tell
+    tell newSession
+        write text "{}"
+    end tell
+end tell
+""".format(argv[-1])
+        with tempfile.NamedTemporaryFile(delete=False, mode='wt+') as tmp:
+            tmp.write(osa_script.lstrip())
+            tmp.flush()
+            os.chmod(tmp.name, 0o700)
+            argv = [which(terminal), tmp.name]
+
     log.debug("Launching a new terminal: %r" % argv)
 
     stdin = stdout = stderr = open(os.devnull, 'r+b')
@@ -406,7 +450,12 @@ os.execve({argv0!r}, {argv!r}, os.environ)
 
     if terminal == 'tmux':
         out, _ = p.communicate()
-        pid = int(out)
+        try:
+            pid = int(out)
+        except ValueError:
+            pid = None
+        if pid is None:
+            log.error("Could not parse PID from tmux output (%r). Start tmux first.", out)
     elif terminal == 'qdbus':
         with subprocess.Popen((qdbus, konsole_dbus_service, '/Sessions/{}'.format(last_konsole_session),
                                'org.kde.konsole.Session.processId'), stdout=subprocess.PIPE) as proc:
@@ -437,6 +486,7 @@ def parse_ldd_output(output):
       output(str): The output to parse
 
     Example:
+
         >>> sorted(parse_ldd_output('''
         ...     linux-vdso.so.1 =>  (0x00007fffbf5fe000)
         ...     libtinfo.so.5 => /lib/x86_64-linux-gnu/libtinfo.so.5 (0x00007fe28117f000)
@@ -501,6 +551,7 @@ def register_sizes(regs, in_sizes):
     Used in i386/AMD64 shellcode, e.g. the mov-shellcode.
 
     Example:
+
         >>> regs = [['eax', 'ax', 'al', 'ah'],['ebx', 'bx', 'bl', 'bh'],
         ... ['ecx', 'cx', 'cl', 'ch'],
         ... ['edx', 'dx', 'dl', 'dh'],
@@ -612,3 +663,239 @@ def python_2_bytes_compatible(klass):
         if '__str__' not in klass.__dict__:
             klass.__str__ = klass.__bytes__
     return klass
+
+def _create_execve_script(argv=None, executable=None, cwd=None, env=None, ignore_environ=None,
+        stdin=0, stdout=1, stderr=2, preexec_fn=None, preexec_args=(), aslr=None, setuid=None,
+        shell=False, log=log):
+    """
+    Creates a python wrapper script that triggers the syscall `execve` directly.
+
+    Arguments:
+        argv(list):
+            List of arguments to pass into the process
+        executable(str):
+            Path to the executable to run.
+            If :const:`None`, ``argv[0]`` is used.
+        cwd(str):
+            Working directory.  If :const:`None`, uses the working directory specified
+            on :attr:`cwd` or set via :meth:`set_working_directory`.
+        env(dict):
+            Environment variables to add to the environment.
+        ignore_environ(bool):
+            Ignore default environment.  By default use default environment iff env not specified.
+        stdin(int, str):
+            If an integer, replace stdin with the numbered file descriptor.
+            If a string, a open a file with the specified path and replace
+            stdin with its file descriptor.  May also be one of ``sys.stdin``,
+            ``sys.stdout``, ``sys.stderr``.  If :const:`None`, the file descriptor is closed.
+        stdout(int, str):
+            See ``stdin``.
+        stderr(int, str):
+            See ``stdin``.
+        preexec_fn(callable):
+            Function which is executed on the remote side before execve().
+            This **MUST** be a self-contained function -- it must perform
+            all of its own imports, and cannot refer to variables outside
+            its scope.
+        preexec_args(object):
+            Argument passed to ``preexec_fn``.
+            This **MUST** only consist of native Python objects.
+        aslr(bool):
+            See :class:`pwnlib.tubes.process.process` for more information.
+        setuid(bool):
+            See :class:`pwnlib.tubes.process.process` for more information.
+        shell(bool):
+            Pass the command-line arguments to the shell.
+
+    Returns:
+        A string containing the python script.
+    """
+    if not argv and not executable:
+        log.error("Must specify argv or executable")
+
+    aslr      = aslr if aslr is not None else context.aslr
+
+    if ignore_environ is None:
+        ignore_environ = env is not None  # compat
+
+    argv, env = normalize_argv_env(argv, env, log)
+
+    if shell:
+        if len(argv) != 1:
+            log.error('Cannot provide more than 1 argument if shell=True')
+        argv = [bytearray(b'/bin/sh'), bytearray(b'-c')] + argv
+
+    executable = executable or argv[0]
+    cwd        = cwd or '.'
+
+    # Validate, since failures on the remote side will suck.
+    if not isinstance(executable, (six.text_type, six.binary_type, bytearray)):
+        log.error("executable / argv[0] must be a string: %r" % executable)
+    executable = bytearray(packing._need_bytes(executable, min_wrong=0x80))
+
+    # Allow passing in sys.stdin/stdout/stderr objects
+    handles = {sys.stdin: 0, sys.stdout:1, sys.stderr:2}
+    stdin  = handles.get(stdin, stdin)
+    stdout = handles.get(stdout, stdout)
+    stderr = handles.get(stderr, stderr)
+
+    # Allow the user to provide a self-contained function to run
+    def func(): pass
+    func      = preexec_fn or func
+    func_args = preexec_args
+
+    if not isinstance(func, types.FunctionType):
+        log.error("preexec_fn must be a function")
+
+    func_name = func.__name__
+    if func_name == (lambda: 0).__name__:
+        log.error("preexec_fn cannot be a lambda")
+
+    func_src  = inspect.getsource(func).strip()
+    setuid = True if setuid is None else bool(setuid)
+
+
+    script = r"""
+#!/usr/bin/env python
+import os, sys, ctypes, resource, platform, stat
+from collections import OrderedDict
+try:
+    integer_types = int, long
+except NameError:
+    integer_types = int,
+exe   = bytes(%(executable)r)
+argv  = [bytes(a) for a in %(argv)r]
+env   = %(env)r
+
+os.chdir(%(cwd)r)
+
+if %(ignore_environ)r:
+    os.environ.clear()
+
+environ = getattr(os, 'environb', os.environ)
+
+if env is not None:
+    env = OrderedDict((bytes(k), bytes(v)) for k,v in env)
+    environ.update(env)
+else:
+    env = environ
+
+def is_exe(path):
+    return os.path.isfile(path) and os.access(path, os.X_OK)
+
+PATH = environ.get(b'PATH',b'').split(os.pathsep.encode())
+
+if os.path.sep.encode() not in exe and not is_exe(exe):
+    for path in PATH:
+        test_path = os.path.join(path, exe)
+        if is_exe(test_path):
+            exe = test_path
+            break
+
+if not is_exe(exe):
+    sys.stderr.write('3\n')
+    sys.stderr.write("{!r} is not executable or does not exist in $PATH: {!r}".format(exe,PATH))
+    sys.exit(-1)
+
+if not %(setuid)r:
+    PR_SET_NO_NEW_PRIVS = 38
+    result = ctypes.CDLL('libc.so.6').prctl(PR_SET_NO_NEW_PRIVS, 1, 0, 0, 0)
+
+    if result != 0:
+        sys.stdout.write('3\n')
+        sys.stdout.write("Could not disable setuid: prctl(PR_SET_NO_NEW_PRIVS) failed")
+        sys.exit(-1)
+
+try:
+    PR_SET_PTRACER = 0x59616d61
+    PR_SET_PTRACER_ANY = -1
+    ctypes.CDLL('libc.so.6').prctl(PR_SET_PTRACER, PR_SET_PTRACER_ANY, 0, 0, 0)
+except Exception:
+    pass
+
+# Determine what UID the process will execute as
+# This is used for locating apport core dumps
+suid = os.getuid()
+sgid = os.getgid()
+st = os.stat(exe)
+if %(setuid)r:
+    if (st.st_mode & stat.S_ISUID):
+        suid = st.st_uid
+    if (st.st_mode & stat.S_ISGID):
+        sgid = st.st_gid
+
+if sys.argv[-1] == 'check':
+    sys.stdout.write("1\n")
+    sys.stdout.write(str(os.getpid()) + "\n")
+    sys.stdout.write(str(os.getuid()) + "\n")
+    sys.stdout.write(str(os.getgid()) + "\n")
+    sys.stdout.write(str(suid) + "\n")
+    sys.stdout.write(str(sgid) + "\n")
+    getattr(sys.stdout, 'buffer', sys.stdout).write(os.path.realpath(exe) + b'\x00')
+    sys.stdout.flush()
+
+for fd, newfd in {0: %(stdin)r, 1: %(stdout)r, 2:%(stderr)r}.items():
+    if newfd is None:
+        os.close(fd)
+    elif isinstance(newfd, (str, bytes)):
+        newfd = os.open(newfd, os.O_RDONLY if fd == 0 else (os.O_RDWR|os.O_CREAT))
+        os.dup2(newfd, fd)
+        os.close(newfd)
+    elif isinstance(newfd, integer_types) and newfd != fd:
+        os.dup2(fd, newfd)
+
+if not %(aslr)r:
+    if platform.system().lower() == 'linux' and %(setuid)r is not True:
+        ADDR_NO_RANDOMIZE = 0x0040000
+        ctypes.CDLL('libc.so.6').personality(ADDR_NO_RANDOMIZE)
+
+    resource.setrlimit(resource.RLIMIT_STACK, (-1, -1))
+
+# Attempt to dump ALL core file regions
+try:
+    with open('/proc/self/coredump_filter', 'w') as core_filter:
+        core_filter.write('0x3f\n')
+except Exception:
+    pass
+
+# Assume that the user would prefer to have core dumps.
+try:
+    resource.setrlimit(resource.RLIMIT_CORE, (-1, -1))
+except Exception:
+    pass
+
+%(func_src)s
+%(func_name)s(*%(func_args)r)
+
+""" % locals()  
+
+    if len(argv) > 0 and len(argv[0]) > 0:
+        script += r"os.execve(exe, argv, env) " 
+
+    # os.execve does not allow us to pass empty argv[0]
+    # Therefore we use ctypes to call execve directly
+    else:
+        script += r"""
+# Transform envp from dict to list
+env_list = [key + b"=" + value for key, value in env.items()]
+
+# ctypes helper to convert a python list to a NULL-terminated C array
+def to_carray(py_list):
+    py_list += [None] # NULL-terminated
+    return (ctypes.c_char_p * len(py_list))(*py_list)
+
+c_argv = to_carray(argv)
+c_env = to_carray(env_list)
+
+# Call execve
+libc = ctypes.CDLL('libc.so.6')
+libc.execve(exe, c_argv, c_env)
+
+# We should never get here, since we sanitized argv and env,
+# but just in case, indicate that something went wrong.
+libc.perror(b"execve")
+raise OSError("execve failed")
+""" % locals()
+    script = script.strip()
+
+    return script
